from collections.abc import Iterable as IterableABC
from email import headerregistry as hr
from email.message import EmailMessage
from mimetypes import guess_type
import os
from typing import Any, Dict, Iterable, List, Tuple, Union
import attr

AnyPath = Union[bytes, str, "os.PathLike[bytes]", "os.PathLike[str]"]

SingleAddress = Union[str, hr.Address]

AddressOrGroup = Union[str, hr.Address, hr.Group]


@attr.s(auto_attribs=True)
class ContentType:
    maintype: str
    subtype: str
    params: Dict[str, Any]

    @classmethod
    def parse(cls, s: str) -> "ContentType":
        maintype, subtype, params = parse_content_type(s)
        return cls(maintype, subtype, params)


def compile_address(addr: SingleAddress) -> hr.Address:
    if isinstance(addr, str):
        return hr.Address(addr_spec=addr)
    else:
        return addr


def compile_addresses(
    addrs: Union[AddressOrGroup, Iterable[AddressOrGroup]]
) -> List[Union[hr.Address, hr.Group]]:
    if isinstance(addrs, str):
        return [hr.Address(addr_spec=addrs)]
    elif not isinstance(addrs, IterableABC):
        return [addrs]
    else:
        return [hr.Address(addr_spec=a) if isinstance(a, str) else a for a in addrs]


def parse_content_type(s: str) -> Tuple[str, str, Dict[str, Any]]:
    """
    Split a :mailheader:`Content-Type` value into a triple of ``(maintype,
    subtype, params)``
    """
    msg = EmailMessage()
    msg["Content-Type"] = s
    if msg["Content-Type"].defects:
        raise ValueError(s)
    ct = msg["Content-Type"]
    return (ct.maintype, ct.subtype, dict(ct.params))


def get_mime_type(filename: str, strict: bool = False) -> str:
    """
    Like `mimetypes.guess_type()`, except that if the file is compressed, the
    MIME type for the compression is returned.  Also, the default return value
    is now ``'application/octet-stream'`` instead of `None`.
    """
    mtype, encoding = guess_type(filename, strict)
    if encoding is None:
        return mtype or "application/octet-stream"
    elif encoding == "gzip":
        # application/gzip is defined by RFC 6713
        return "application/gzip"
        # There is also a "+gzip" MIME structured syntax suffix defined by RFC
        # 8460; exactly when can that be used?
        # return mtype + '+gzip'
    else:
        return "application/x-" + encoding



from typing import Any, Dict, List, Tuple, Union
import pytest
from eletter import (
    Address,
    Group,
    assemble_content_type,
    format_addresses,
    reply_quote,
)
from eletter.util import get_mime_type, parse_content_type


@pytest.mark.parametrize(
    "s,ct",
    [
        ("text/plain", ("text", "plain", {})),
        ("TEXT/PLAIN", ("text", "plain", {})),
        ("text/plain; charset=utf-8", ("text", "plain", {"charset": "utf-8"})),
        ('text/plain; charset="utf-8"', ("text", "plain", {"charset": "utf-8"})),
        (
            "text/markdown; charset=utf-8; variant=GFM",
            ("text", "markdown", {"charset": "utf-8", "variant": "GFM"}),
        ),
        (
            'text/plain; charset="utf-\u2603"',
            ("text", "plain", {"charset": "utf-\u2603"}),
        ),
        (
            "text/plain; charset*=utf-8''utf-%E2%98%83",
            ("text", "plain", {"charset": "utf-\u2603"}),
        ),
    ],
)
def test_parse_content_type(s: str, ct: Tuple[str, str, Dict[str, Any]]) -> None:
    assert parse_content_type(s) == ct


@pytest.mark.parametrize(
    "s",
    [
        "text",
        "text/",
        "/plain",
        "text/plain, charset=utf-8",
    ],
)
def test_parse_content_type_error(s: str) -> None:
    with pytest.raises(ValueError):
        parse_content_type(s)


@pytest.mark.parametrize(
    "maintype,subtype,params,ct",
    [
        ("text", "plain", {}, "text/plain"),
        ("TEXT", "PLAIN", {}, "TEXT/PLAIN"),
        ("text", "plain", {"charset": "utf-8"}, 'text/plain; charset="utf-8"'),
        ("text", "plain", {"name": "résumé.txt"}, 'text/plain; name="résumé.txt"'),
        ("text", "plain", {"name": 'foo"bar'}, 'text/plain; name="foo\\"bar"'),
        (
            "text",
            "markdown",
            {"charset": "utf-8", "variant": "GFM"},
            'text/markdown; charset="utf-8"; variant="GFM"',
        ),
    ],
)
def test_assemble_content_type(
    maintype: str, subtype: str, params: Dict[str, str], ct: str
) -> None:
    assert assemble_content_type(maintype, subtype, **params) == ct


@pytest.mark.parametrize(
    "maintype,subtype",
    [
        ("text/plain", "plain"),
        ("text", ""),
        ("text/", "plain"),
    ],
)
def test_assemble_content_type_error(maintype: str, subtype: str) -> None:
    with pytest.raises(ValueError) as excinfo:
        assemble_content_type(maintype, subtype)
    assert str(excinfo.value) == f"{maintype}/{subtype}"


@pytest.mark.parametrize(
    "filename,mtype",
    [
        ("foo.txt", "text/plain"),
        ("foo", "application/octet-stream"),
        ("foo.gz", "application/gzip"),
        ("foo.tar.gz", "application/gzip"),
        ("foo.tgz", "application/gzip"),
        ("foo.taz", "application/gzip"),
        ("foo.svg.gz", "application/gzip"),
        ("foo.svgz", "application/gzip"),
        ("foo.Z", "application/x-compress"),
        ("foo.tar.Z", "application/x-compress"),
        ("foo.bz2", "application/x-bzip2"),
        ("foo.tar.bz2", "application/x-bzip2"),
        ("foo.tbz2", "application/x-bzip2"),
        ("foo.xz", "application/x-xz"),
        ("foo.tar.xz", "application/x-xz"),
        ("foo.txz", "application/x-xz"),
    ],
)
def test_get_mime_type(filename: str, mtype: str) -> None:
    assert get_mime_type(filename) == mtype


@pytest.mark.parametrize(
    "inp,output",
    [
        ("", "> \n"),
        ("\n", "> \n"),
        ("Insert output here.", "> Insert output here.\n"),
        ("Insert output here.\n", "> Insert output here.\n"),
        (
            "Insert output here.\nOutsert input there.",
            "> Insert output here.\n> Outsert input there.\n",
        ),
        (
            "Insert output here.\nOutsert input there.\n",
            "> Insert output here.\n> Outsert input there.\n",
        ),
        (
            "Insert output here.\r\nOutsert input there.\r\n",
            "> Insert output here.\r\n> Outsert input there.\r\n",
        ),
        (
            "Insert output here.\rOutsert input there.\r",
            "> Insert output here.\r> Outsert input there.\r",
        ),
        (
            "Insert output here.\n\nOutsert input there.\n",
            "> Insert output here.\n> \n> Outsert input there.\n",
        ),
        (
            "> Insert output here.\n> \n> Outsert input there.\n",
            ">> Insert output here.\n>> \n>> Outsert input there.\n",
        ),
    ],
)
def test_reply_quote(inp: str, output: str) -> None:
    assert reply_quote(inp) == output


def test_reply_quote_custom_prefix() -> None:
    assert (
        reply_quote("Insert output here.\n\n: Outsert input there.\n", ": ")
        == ": Insert output here.\n: \n:: Outsert input there.\n"
    )


@pytest.mark.parametrize(
    "addresses,fmt",
    [
        ([], ""),
        (["foo@example.com"], "foo@example.com"),
        (["foo@example.com", "bar@example.org"], "foo@example.com, bar@example.org"),
        ([Address("Fabian Oo", "foo@example.com")], "Fabian Oo <foo@example.com>"),
        (
            [
                Address("Fabian Oo", "foo@example.com"),
                Address("Bastian Arrr", "bar@example.org"),
            ],
            "Fabian Oo <foo@example.com>, Bastian Arrr <bar@example.org>",
        ),
        (
            [Address("Fabian O. Oh", "foo@example.com")],
            '"Fabian O. Oh" <foo@example.com>',
        ),
        (
            [Address("Zoë Façade", "zoe.facade@naïveté.fr")],
            "Zoë Façade <zoe.facade@naïveté.fr>",
        ),
        (
            [
                Group("undisclosed recipients", []),
                "luser@example.nil",
                Group(
                    "friends",
                    ["you@there.net", Address("Thaddeus Hem", "them@hither.yon")],
                ),
            ],
            "undisclosed recipients:;, luser@example.nil,"
            " friends: you@there.net, Thaddeus Hem <them@hither.yon>;",
        ),
    ],
)
def test_format_addresses(
    addresses: List[Union[str, Address, Group]], fmt: str
) -> None:
    assert format_addresses(addresses) == fmt



from datetime import datetime
from email.message import EmailMessage
from typing import Iterable, Mapping, Optional, Union
from .classes import Attachment, HTMLBody, MailItem, TextBody
from .util import AddressOrGroup, SingleAddress, compile_addresses


def compose(
    subject: str,
    from_: Union[AddressOrGroup, Iterable[AddressOrGroup]],
    to: Iterable[AddressOrGroup],
    text: Optional[str] = None,
    html: Optional[str] = None,
    cc: Optional[Iterable[AddressOrGroup]] = None,
    bcc: Optional[Iterable[AddressOrGroup]] = None,
    reply_to: Optional[Union[AddressOrGroup, Iterable[AddressOrGroup]]] = None,
    sender: Optional[SingleAddress] = None,
    date: Optional[datetime] = None,
    headers: Optional[Mapping[str, Union[str, Iterable[str]]]] = None,
    attachments: Optional[Iterable[Attachment]] = None,
) -> EmailMessage:
    """
    Construct an `~email.message.EmailMessage` instance from a subject,
    :mailheader:`From` address, :mailheader:`To` addresses, and a plain text
    and/or HTML body, optionally accompanied by attachments and other headers.

    :param str subject: The e-mail's :mailheader:`Subject` line
    :param from_:
        The e-mail's :mailheader:`From` line.  Note that this argument is
        spelled with an underscore, as "``from``" is a keyword in Python.
    :type from_: address or iterable of addresses
    :param to: The e-mail's :mailheader:`To` line
    :type to: iterable of addresses
    :param str text:
        The contents of a :mimetype:`text/plain` body for the e-mail.  At least
        one of ``text`` and ``html`` must be specified.
    :param str html:
        The contents of a :mimetype:`text/html` body for the e-mail.  At least
        one of ``text`` and ``html`` must be specified.
    :param cc: The e-mail's :mailheader:`CC` line
    :type cc: iterable of addresses
    :param bcc: The e-mail's :mailheader:`BCC` line
    :type bcc: iterable of addresses
    :param reply_to: The e-mail's :mailheader:`Reply-To` line
    :type reply_to: address or iterable of addresses
    :param address sender:
        The e-mail's :mailheader:`Sender` line.  The address must be a string
        or `Address`, not a `Group`.
    :param datetime date: The e-mail's :mailheader:`Date` line
    :param attachments:
        A collection of :ref:`attachments <attachments>` to append to the
        e-mail
    :type attachments: iterable of attachments
    :param mapping headers:
        A collection of additional headers to add to the e-mail.  A header
        value may be either a single string or an iterable of strings to add
        multiple headers with the same name.  If you wish to set an
        otherwise-unsupported address header like :mailheader:`Resent-From` to
        a list of addresses, use the `format_addresses()` function to first
        convert the addresses to a string.
    :rtype: email.message.EmailMessage
    :raises ValueError: if neither ``text`` nor ``html`` is set
    """
    msg: Optional[MailItem] = None
    if text is not None:
        msg = TextBody(text)
    if html is not None:
        if msg is None:
            msg = HTMLBody(html)
        else:
            msg |= HTMLBody(html)
    if msg is None:
        raise ValueError("At least one of text and html must be non-None")
    if attachments is not None:
        for a in attachments:
            msg &= a
    return msg.compose(
        subject=subject,
        from_=from_,
        to=to,
        cc=cc,
        bcc=bcc,
        reply_to=reply_to,
        sender=sender,
        date=date,
        headers=headers,
    )


def assemble_content_type(maintype: str, subtype: str, **params: str) -> str:
    """
    Construct a :mailheader:`Content-Type` string from a maintype, subtype, and
    some number of parameters

    :raises ValueError: if ``f"{maintype}/{subtype}"`` is an invalid
        :mailheader:`Content-Type`
    """
    ct = f"{maintype}/{subtype}"
    msg = EmailMessage()
    msg["Content-Type"] = ct
    if msg["Content-Type"].defects:
        raise ValueError(ct)
    for k, v in params.items():
        msg.set_param(k, v)
    return str(msg["Content-Type"])


def reply_quote(s: str, prefix: str = "> ") -> str:
    """
    Quote__ a text following the *de facto* standard for replying to an e-mail;
    that is, prefix each line of the text with ``"> "`` (or a custom prefix),
    and if a line already starts with the prefix, omit any trailing whitespace
    from the newly-added prefix (so ``"> already quoted"`` becomes ``">>
    already quoted"``).

    If the resulting string does not end with a newline, one is added.  The
    empty string is treated as a single line.

    __ https://en.wikipedia.org/wiki/Usenet_quoting
    """
    s2 = ""
    for ln in (s or "\n").splitlines(True):
        if ln.startswith(prefix):
            s2 += prefix.rstrip() + ln
        else:
            s2 += prefix + ln
    if not s2.endswith(("\n", "\r")):
        s2 += "\n"
    return s2


def format_addresses(addresses: Iterable[AddressOrGroup]) -> str:
    """
    Format a sequence of addresses for use in a custom address header string
    """
    msg = EmailMessage()
    msg["To"] = compile_addresses(addresses)
    return str(msg["To"])
